<?php

namespace Drupal\commerce_shipstation\Controller;

use Drupal\commerce_shipstation\ShipStation;
use Drupal\Core\Controller\ControllerBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\AccessDeniedHttpException;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;

/**
 *
 */
class ShipStationEndpointController extends ControllerBase {

  /**
   * ShipStation Configuration.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $shipStationConfig;

  /**
   * ShipStation Service.
   *
   * @var \Drupal\commerce_shipstation\ShipStation
   */
  protected $shipStation;

  /**
   * Constructs a new controller.
   *
   * @param \Drupal\commerce_shipstation\ShipStation $shipStationService
   *   Constructs a new ShipStation Service.
   */
  public function __construct(ShipStation $shipstation_service) {
    $this->shipStationConfig = $this->config('commerce_shipstation.shipstation_config');
    $this->shipStation = $shipstation_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('commerce_shipstation.shipstation_service')
    );
  }

  /**
   * Establish a service endpoint for shipstation to communicate with.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The response.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   */
  public function shipStationEndpointRequest(Request $request) {
    \Drupal::service('page_cache_kill_switch')->trigger();
    $logging = $this->shipStationConfig->get('commerce_shipstation_logging');
    // ShipStation starts at 1, Drupal at 0.
    $request->query->set('page', (int) $request->query->get('page') - 1);

    // Log each request to the endpoint if logging is enabled.
    if ($logging) {
      $request_vars = $request->query->all();
      // Obfuscate the sensitive data before logging the request.
      $request_vars['SS-UserName'] = '******';
      $request_vars['SS-Password'] = '******';
      // $request_vars['auth_key'] = '*****';
      \Drupal::logger('commerce_shipstation')->info('ShipStation request: @get', ['@get' => var_export($request_vars, TRUE)]);
    }

    // Run the call based on the action it defines.
    $response = new Response();
    switch ($request->query->get('action')) {
      case ShipStation::EXPORT_ACTION:
        $start_date = $request->query->get('start_date') ?: '-1 day';
        $end_date = $request->query->get('end_date') ?: 'now';
        $page = $request->query->get('page') ?: 0;
        $xml = $this->shipStation->exportOrders($start_date, $end_date, $page);

        $response->headers->set('Content-type', 'application/xml');
        $response->setContent($xml);
        return $response;

      case ShipStation::SHIPNOTIFY_ACTION:
        $order_number = $request->query->get('order_number');
        $tracking_number = $request->query->get('tracking_number');
        $carrier = $request->query->get('carrier');
        $ship_date = $request->query->get('ship_date');
        $msg = $this->shipStation->requestShipNotify($order_number, $tracking_number, $carrier, $ship_date);
        $response->setContent($msg);
        return $response;

      default:
        \Drupal::messenger()->addError(t('The ShipStation request action is invalid'));
        \Drupal::logger('commerce_shipstation')->error('Invalid request action received from ShipStation. Enable or check request logging for more information');
        throw new AccessDeniedHttpException();
    }
  }

}
